@@ -14,7 +14,7 @@ except ImportError:
 from txpostgres import txpostgres
 
 from twisted.trial import unittest
-from twisted.internet import defer, reactor
+from twisted.internet import defer, posixbase, reactor
 
 simple_table_schema = """
 CREATE TABLE simple (
@@ -92,12 +92,16 @@ class FakeReactor(object):
     def reset(self):
         self.readersAdded = self.writersAdded = 0
         self.readersRemoved = self.writersRemoved = 0
+
     def addReader(self, _):
         self.readersAdded += 1
+
     def addWriter(self, _):
         self.writersAdded += 1
+
     def removeReader(self, _):
         self.readersRemoved += 1
+
     def removeWriter(self, _):
         self.writersRemoved += 1
 
@@ -182,17 +186,16 @@ class TxPostgresPollingMixinTestCase(Psycopg2TestCase):
 
     def test_connectionLost(self):
         """
-        Calling connectionLost() errbacks the C{Deferred} returned from poll()
-        but another connectionLost() is harmless.
+        Calls to connectionLost() get swallowed.
         """
         p = FakeWrapper()
         p._pollable = PollableThing()
+        p._pollable.NEXT_STATE = psycopg2.extensions.POLL_OK
 
         d = p.poll()
         p.connectionLost(RuntimeError("boom"))
-        d = self.assertFailure(d, RuntimeError)
         p.connectionLost(RuntimeError("bam"))
-        return d
+        return d.addCallback(self.assertEquals, p)
 
     def test_errors(self):
         """
@@ -289,13 +292,18 @@ class TxPostgresConnectionTestCase(Psycopg2TestCase):
         also make connect() return a failure.
         """
         conn = txpostgres.Connection()
+
         class BadPollable(object):
+
             def __init__(*args, **kwars):
                 pass
+
             def poll(self):
                 raise RuntimeError("booga")
+
             def close(self):
                 pass
+
         conn.connectionFactory = BadPollable
 
         d = conn.connect()
@@ -303,22 +311,29 @@ class TxPostgresConnectionTestCase(Psycopg2TestCase):
         d.addCallback(lambda _: conn.close())
 
         class BadThing(object):
+
             def __init__(*args, **kwargs):
                 raise RuntimeError("wooga")
+
             def close(self):
                 pass
+
         conn.connectionFactory = BadThing
 
         d.addCallback(lambda _: conn.connect())
         d = self.assertFailure(d, RuntimeError)
 
         class BrokenPollable(object):
+
             def __init__(*args, **kwars):
                 pass
+
             def poll(self):
                 return "tee hee hee"
+
             def close(self):
                 pass
+
         conn.connectionFactory = BrokenPollable
 
         d.addCallback(lambda _: conn.connect())
@@ -560,6 +575,72 @@ class TxPostgresQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         # rollback for real, or tearDown won't be able to drop the table
         return d.addCallback(lambda _: self.conn.runOperation("rollback"))
 
+    def test_terminatedConnection(self):
+        """
+        If the connection gets terminated (because of a segmentation fault,
+        administrative backend termination or other circumstances), a failure
+        wrapping the original psycopg2 error is returned and subsequent queries
+        fail with an error indicating that the connection is already closed.
+        """
+        # this tests uses pg_terminate_backend, so it only works on PostgreSQL
+        # 8.4+ and if the user running the tests is a superuser.
+        if self.conn.server_version < 84000:
+            raise unittest.SkipTest(
+                "PostgreSQL < 8.4.0 does not have pg_terminate_backend")
+
+        # check if this Twisted has a patch for #4539, otherwise the test will
+        # fail because the terminated cursor will have fileno() called on it
+        if not getattr(posixbase, '_PollLikeMixin', None):
+            raise unittest.SkipTest("This test fails on versions of Twisted "
+                                    "affected by Twisted bug #4539")
+
+        def checkSuperuser(ret):
+            if ret[0][0] != 'on':
+                raise unittest.SkipTest(
+                    "This test uses pg_terminate_backend, "
+                    "which can only be called by a database superuser")
+
+        d = self.conn.runQuery("show is_superuser")
+        d.addCallback(checkSuperuser)
+
+        def terminateAndRunQuery():
+            d = self.conn.runQuery("select pg_terminate_backend(%s)",
+                                   (self.conn.get_backend_pid(), ))
+            d = self.assertFailure(d, psycopg2.DatabaseError)
+            d.addCallback(lambda _: self.conn.runQuery("select 1"))
+            d = self.assertFailure(d, psycopg2.InterfaceError)
+
+            def restoreConnection():
+                self.conn = txpostgres.Connection()
+                return self.conn.connect(user=DB_USER, password=DB_PASS,
+                                         host=DB_HOST, database=DB_NAME)
+
+            # restore the connection, otherwise all the other tests will fail
+            return d.addCallback(lambda _: restoreConnection())
+
+        return d.addCallback(lambda _: terminateAndRunQuery())
+
+    def test_connectionLostWhileRunning(self):
+        cursors = []
+
+        class RetainingCursor(txpostgres.Cursor):
+
+            def __init__(self, cursor, connection):
+                cursors.append(self)
+                txpostgres.Cursor.__init__(self, cursor, connection)
+
+        mp = self.patch(self.conn, 'cursorFactory', RetainingCursor)
+
+        d1 = self.conn.runQuery("select 1")
+        d2 = self.conn.runQuery("select 1")
+
+        self.assertEquals(len(cursors), 1)
+        cursors[0].connectionLost(RuntimeError("boom"))
+
+        d = defer.gatherResults([d1, d2])
+        d.addCallback(self.assertEquals, [[(1, )], [(1, )]])
+        return d.addCallback(lambda _: mp.restore())
+
 
 class TxPostgresConnectionPoolTestCase(Psycopg2TestCase):
 
@@ -654,6 +735,7 @@ class TxPostgresConnectionPoolHotswappingTestCase(Psycopg2TestCase):
             return c.execute("boom")
         d.addCallback(lambda _: pool.runInteraction(brokenInteraction))
         d.addCallback(lambda _: self.fail("No exception"))
+
         def checkErrorAndHotswap(f):
             f.trap(txpostgres.RollbackFailed)
             e = f.value
@@ -670,6 +752,7 @@ class TxPostgresConnectionPoolHotswappingTestCase(Psycopg2TestCase):
             d = c.connect(user=DB_USER, password=DB_PASS,
                           host=DB_HOST, database=DB_NAME)
             return d.addCallback(lambda c: pool.add(c))
+
         d.addErrback(checkErrorAndHotswap)
 
         d.addCallback(lambda _: defer.gatherResults([
