@@ -8,10 +8,9 @@ driver.
 """
 
 import psycopg2
-from psycopg2 import extensions
 from zope.interface import implements
 
-from twisted.internet import interfaces, reactor, defer
+from twisted.internet import interfaces, defer
 from twisted.python import failure, log
 
 
@@ -30,6 +29,7 @@ class UnexpectedPollResult(Exception):
     Polling returned an unexpected result.
     """
 
+
 class _CancelInProgress(Exception):
     """
     A query cancellation is in progress.
@@ -119,9 +119,12 @@ class _PollingMixin(object):
             return -1
 
     def connectionLost(self, reason):
-        if self._pollingD:
-            d, self._pollingD = self._pollingD, None
-            d.errback(reason)
+        # Do not errback self._pollingD here! We need to keep on calling poll()
+        # until it reports an error, which will errback self._pollingD with the
+        # correct failure. If we errback here, we won't finish the poll()
+        # cycle, which would leave psycopg2 in a state where it thinks there's
+        # still an async query underway.
+        return
 
     def _cancel(self, d):
         try:
@@ -330,7 +333,7 @@ class Connection(_PollingMixin):
         c = self.cursor()
         d = c.execute(*args, **kwargs)
         d.addCallback(lambda c: c.fetchall())
-        return d.addBoth(lambda ret: (c.close(), ret)[1])
+        return d.addCallback(lambda ret: (c.close(), ret)[1])
 
     def runOperation(self, *args, **kwargs):
         """
@@ -352,7 +355,7 @@ class Connection(_PollingMixin):
         c = self.cursor()
         d = c.execute(*args, **kwargs)
         d.addCallback(lambda _: None)
-        return d.addBoth(lambda ret: (c.close(), ret)[1])
+        return d.addCallback(lambda ret: (c.close(), ret)[1])
 
     def runInteraction(self, interaction, *args, **kwargs):
         """
@@ -401,20 +404,24 @@ class Connection(_PollingMixin):
         def commitAndPassthrough(ret, cursor):
             e = cursor.execute("commit")
             return e.addCallback(lambda _: ret)
+
         def rollbackAndPassthrough(f, cursor):
             # maybeDeferred in case cursor.execute raises a synchronous
             # exception
             e = defer.maybeDeferred(cursor.execute, "rollback")
+
             def justPanic(rf):
                 log.err(rf)
                 return defer.fail(RollbackFailed(self, f))
+
             # if rollback failed panic
             e.addErrback(justPanic)
             # reraise the original failure afterwards
             return e.addCallback(lambda _: f)
+
         d.addCallback(commitAndPassthrough, c)
         d.addErrback(rollbackAndPassthrough, c)
-        d.addBoth(lambda ret: (c.close(), ret)[1])
+        d.addCallback(lambda ret: (c.close(), ret)[1])
 
         return d
 
@@ -530,7 +537,7 @@ class ConnectionPool(object):
         """
         self.connections.add(connection)
         self._semaphore.limit += 1
-        self._semaphore.release() # uuuugh...
+        self._semaphore.release()  # uuuugh...
 
     def _putBackAndPassthrough(self, result, connection):
         self.connections.add(connection)
